program create
  ! Create eam/fs interatomic potential file for LAMMPS from parameters given in EAM.input.
  ! Author: Jyri Kimari jyri.kimari@helsinki.fi, 2023-05-08.
  implicit none

  integer :: n_elements,VAA_knots,VAB_knots,PhiAA_knots

  character(len=80) :: comment,citation

  integer :: Nrho,Nr
  double precision :: drho,dr,cutoff
  
  character(len=2),allocatable,dimension(:) :: element_names
  integer,allocatable,dimension(:) :: atomic_numbers
  double precision,allocatable,dimension(:) :: masses,lattice_constants
  character(len=3),allocatable,dimension(:) :: lattice_types

  double precision,allocatable,dimension(:,:,:) :: rAB,aAB
  double precision,allocatable,dimension(:,:) :: rAA,aAA
  double precision,allocatable,dimension(:,:) :: RPhi,APhi

  integer :: A
  character(len=80) :: filename

  write(6,"(a)") "Reading parameters from EAM.input"
  call read_input()

  write(6,"(a)") "Writing F, Phi and V to F.dat, Phi.dat and V.dat"
  call write_functions()

  filename=""
  do A=1,n_elements
    filename=trim(filename)//element_names(A)
  end do
  filename=trim(filename)//".eam.fs"
  write(6,"(a)") "Writing potential to "//trim(filename) 
  call write_setfl()
  
  write(6,"(a)") "Finished!"

contains

  integer function H(x)
    ! Heaviside function
    double precision,intent(in) :: x

    if (x>=0) then
      H=1
    else
      H=0
    endif
  end function H

  double precision function V(A,B,Rij)
    ! Lazy interface of sorts for pair interaction functions
    integer,intent(in) :: A,B
    double precision,intent(in) :: Rij

    if (A==B) then
      V=VAA(A,Rij)
    else
      V=VAB(A,B,Rij)
    endif
  end function V

  double precision function VAA(A,Rij)
    ! Pair interaction between atoms of same element
    integer,intent(in) :: A
    double precision,intent(in) :: Rij
    
    integer :: k
    
    VAA=0.0d0
    do k=1,VAA_knots
      VAA=VAA + aAA(k,A) * H( rAA(k,A)-Rij ) * ( rAA(k,A)-Rij )**3
    end do

  end function VAA

  double precision function VAB(A,B,Rij)
    ! Pair interaction for different elements
    integer,intent(in) :: A,B
    double precision,intent(in) :: Rij

    integer :: k

    VAB=0.0d0
    do k=1,VAB_knots
      VAB=VAB + aAB(k,A,B) * H( rAB(k,A,B)-Rij ) * ( rAB(k,A,B)-Rij )**3
    end do
  end function VAB

  double precision function Phi(A,B,Rij)
    ! Interface for sum of squares of hopping integrals
    integer,intent(in) :: A,B
    double precision,intent(in) :: Rij

    if (A==B) then
      Phi=PhiAA(A,Rij)
    else
      Phi=PhiAB(A,B,Rij)
    endif
  end function Phi

  double precision function PhiAA(A,Rij)
    ! Sum of squares of hopping integrals, same element
    integer,intent(in) :: A
    double precision,intent(in) :: Rij

    integer :: k

    PhiAA=0.0d0
    do k=1,PhiAA_knots
      PhiAA=PhiAA + APhi(k,A) * H( RPhi(k,A)-Rij ) * ( RPhi(k,A)-Rij )**3
    end do
  end function PhiAA

  double precision function PhiAB(A,B,Rij)
    ! Sum of squares of hopping integrals, different element
    integer,intent(in) :: A,B
    double precision,intent(in) :: Rij

    PhiAB=sqrt(PhiAA(A,Rij)*PhiAA(B,Rij))
  end function PhiAB

  double precision function F(rho)
    ! "Embedding function", just the -sqrt() for this model
    double precision,intent(in) :: rho

    F=-sqrt(rho)
  end function F
  
  subroutine read_input()
    ! Read EAM.input
    integer :: A,B,k,fileunit
    
    open(newunit=fileunit,file="EAM.input",action="read")

    read(fileunit,"(a)") comment
    read(fileunit,"(a)") citation

    read(fileunit,*) n_elements
    read(fileunit,*) VAA_knots
    read(fileunit,*) VAB_knots
    read(fileunit,*) PhiAA_knots

    allocate(rAB(VAB_knots,n_elements,n_elements),aAB(VAB_knots,n_elements,n_elements))
    allocate(rAA(VAA_knots,n_elements),aAA(VAA_knots,n_elements))
    allocate(RPhi(PhiAA_knots,n_elements),APhi(PhiAA_knots,n_elements))
    allocate(element_names(n_elements),atomic_numbers(n_elements),masses(n_elements))
    allocate(lattice_constants(n_elements),lattice_types(n_elements))

    read(fileunit,*) element_names
    read(fileunit,*) atomic_numbers
    read(fileunit,*) masses
    read(fileunit,*) lattice_constants
    read(fileunit,*) lattice_types

    read(fileunit,*) Nrho,drho,Nr,dr,cutoff

    rAB=0.0d0
    do A=1,n_elements-1
      do B=A+1,n_elements
        do k=1,VAB_knots
          read(fileunit,*) rAB(k,A,B)
          ! Symmetry
          rAB(k,B,A)=rAB(k,A,B)
        end do
      end do
    end do

    aAB=0.0d0
    do A=1,n_elements-1
      do B=A+1,n_elements
        do k=1,VAB_knots
          read(fileunit,*) aAB(k,A,B)
          ! Symmetry
          aAB(k,B,A)=aAB(k,A,B)
        end do
      end do
    end do

    rAA=0.0d0
    do A=1,n_elements
      do k=1,VAA_knots
        read(fileunit,*) rAA(k,A)
      end do
      ! Spline knots for self interaction are given in normalized units of lattice constant
      rAA(:,A)=rAA(:,A)*lattice_constants(A)
    end do

    aAA=0.0d0
    do A=1,n_elements
      do k=1,VAA_knots
        read(fileunit,*) aAA(k,A)
      end do
      ! Coefficients for self interaction are given in normalized units of eV, converted to eV/AA^3
      aAA(:,A)=aAA(:,A)/lattice_constants(A)**3
    end do

    RPhi=0.0d0
    do A=1,n_elements
      do k=1,PhiAA_knots
        read(fileunit,*) RPhi(k,A)
      end do
      ! Spline knots for self "density" are given in normalized units of lattice constants
      RPhi(:,A)=RPhi(:,A)*lattice_constants(A)
    end do

    APhi=0.0d0
    do A=1,n_elements
      do k=1,PhiAA_knots
        read(fileunit,*) APhi(k,A)
      end do
      ! Coefficients for self "density" are given in normalized units of eV^2, converted to eV^2/AA^3
      APhi(:,A)=APhi(:,A)/lattice_constants(A)**3
    end do

    close(fileunit)
  end subroutine read_input

  subroutine write_functions()
    ! Write F, V and Phi functions in files to check they seem reasonable.
    integer :: A,B,ir,irho,fileunit

    open(newunit=fileunit,file="F.dat",action="write")
    write(fileunit,"(a)") "i rho F"
    do irho=0,Nrho
      write(fileunit,"(i0,e24.16,e24.16)") irho,irho*drho,F(irho*drho)
    end do
    close(fileunit)

    open(newunit=fileunit,file="Phi.dat",action="write")

    write(fileunit,"(a)",advance="no") "i r"
    do A=1,n_elements
      do B=A,n_elements
        write(fileunit,"(1x,a)",advance="no") element_names(A)//"-"//element_names(B)
      end do
    end do
    write(fileunit,*)

    do ir=0,Nr
      write(fileunit,"(i0,e24.16)",advance="no") ir,ir*dr
      do A=1,n_elements
        do B=A,n_elements
          write(fileunit,"(e24.16)",advance="no") Phi(A,B,ir*dr)
        end do
      end do
      write(fileunit,*)
    end do
    close(fileunit)

    open(newunit=fileunit,file="V.dat",action="write")

    write(fileunit,"(a)",advance="no") "i r"
    do A=1,n_elements
      do B=A,n_elements
        write(fileunit,"(1x,a)",advance="no") element_names(A)//"-"//element_names(B)
      end do
    end do
    write(fileunit,*)

    do ir=0,Nr
      write(fileunit,"(i0,e24.16)",advance="no") ir,ir*dr
      do A=1,n_elements
        do B=A,n_elements
          write(fileunit,"(e24.16)",advance="no") V(A,B,ir*dr)
        end do
      end do
      write(fileunit,*)
    end do
    close(fileunit)
  end subroutine write_functions
  
  subroutine write_setfl()
    ! Write the potential in the setfl/DYNAMO/LAMMPS format.
    integer :: A,B,fileunit
    integer :: i,irho,ir
    integer,dimension(8) :: values
    double precision :: r,rho

    open(newunit=fileunit,file=filename,action="write")
    call date_and_time(values=values)

    write(fileunit,"(a,i0,a,i0.2,a,i0.2,'. ',a)") "DATE: ",values(1),"-",values(2),"-",values(3),comment
    write(fileunit,"(a)") "CITATION: "//trim(citation)
    write(fileunit,"(a)") "Created by create.f90, author Jyri Kimari jyri.kimari@helsinki.fi, 2023-05-08"

    write(fileunit,"(i0)",advance="no") n_elements
    do A=1,n_elements
      write(fileunit,"(1x,a)",advance="no") element_names(A)
    end do
    write(fileunit,*)
    write(fileunit,"(i0,1x,e24.16,1x,i0,1x,e24.16,1x,e24.16)") Nrho,drho,Nr,dr,cutoff

    ! Embedding and density functions
    do A=1,n_elements
      write(fileunit,"(i0,f8.3,f6.3,1x,a)") atomic_numbers(A),masses(A),lattice_constants(A),lattice_types(A)
      ! Embedding function F() is actually the same for all elements in this parametrization,
      ! but the setfl format specification requires it to be repeated under each element.
      ! Furthermore, F() here is merely -sqrt(), but F() is called here for (hopefully) added clarity.
      ! Wrap five values per line for each function
      do irho=0,Nrho-1,5
        do i=0,4
          rho=(irho+i)*drho
          write(fileunit,"(e24.16)",advance="no") F(rho)
        end do
        write(fileunit,*)
      end do
      ! This program assumes symmetric Phi with respect to elements A and B,
      ! but LAMMPS allows asymmetry and requires both Phi(A,B) and Phi(B,A)
      ! to be written
      do B=1,n_elements
        ! Wrap five values per line for each function
        do ir=0,Nr-1,5
          do i=0,4
            r=(ir+i)*dr
            write(fileunit,"(e24.16)",advance="no") Phi(A,B,r)
          end do
          write(fileunit,*)
        end do
      end do
    end do

    ! Pair interactions
    ! Only print pair interactions for A<=B, LAMMPS assumes them symmetric
    do A=1,n_elements
      do B=1,A
        ! Wrap five values per line for each function
        do ir=0,Nr-1,5
          do i=0,4
            r=(ir+i)*dr
            ! Multiply pair interaction function value by r as per the setfl specification
            write(fileunit,"(e24.16)",advance="no") V(A,B,r)*r
          end do
          write(fileunit,*)
        end do
      end do
    end do
    close(fileunit)
  end subroutine write_setfl

end program create
